/*________________________________________________________________________________________________

	Title: Gaussian Blur ROI
	'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Author: Jolyon Troscianko
	Date: 9/10/2018
............................................................................................................

This is essentially a slow & painful Gaussian blur. Unlike a standard Gaussian blur filter, which only
operates on pixel values in the horizontal and vertical planes around the focal pixel, this incorporates
all pixels within the given radius. This feature means it can operate on selected regions which have any
shape, with minimal (or even no) edge-artifact interference on non-rectangular selections

--------------------------------------------------------------------------------------------------------------------------------
Requirements:

Input images should be 32-bit (single or stack). The filter is only applied to the selected area.


_________________________________________________________________________________________________
*/

import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;


public class Gaussian_Blur_ROI implements PlugIn {
public void run(String arg) {


IJ.showStatus("Gaussian Blur ROI");

ImagePlus imp = IJ.getImage();

// ---------------------- Calculate Gaussian Filter Stuff-------------------------------

double sigma = 3.0;

	GenericDialog gd = new GenericDialog("ROI Gaussian Blur");
		gd.addNumericField("Sigma", sigma, 3);
		gd.addMessage("This Gaussian blur works right up to selection edges, but is slow");

	gd.showDialog();
	if (gd.wasCanceled())
		return;

sigma = (double) gd.getNextNumber();


//----------------calculate size of kernel---------------

double gVal = (1/Math.pow(2*Math.PI*Math.pow(sigma,2),0.5)) * Math.exp(-1*(Math.pow(0,2)/(2*Math.pow(sigma,2) ) ));
double minVal = 0.001; // this is the cutoff as a proportion of the peak at which it is assumed the Guassian curve is small enough to have no influence
minVal = minVal / ( (1/Math.pow(2*Math.PI*Math.pow(sigma,2),0.5)) * Math.exp(-1*(Math.pow(0,2)/(2*Math.pow(sigma,2) ) ))); // minVal as a proportaion of start val
int cutoff = 1;

while(gVal > minVal){
	gVal = (1/Math.pow(2*Math.PI*Math.pow(sigma,2),0.5)) * Math.exp(-1*(Math.pow(cutoff,2)/(2*Math.pow(sigma,2) ) ));
	cutoff++;
}

// Gaussian Kernel
int mapW = 2*cutoff +1;
double[] gausMap = new double[mapW*mapW]; 
double rad = 0.0;

for(int y=0; y<mapW; y++)
for(int x=0; x<mapW; x++){
	rad = Math.pow( Math.pow(x-cutoff,2) + Math.pow(y-cutoff,2) ,0.5);
	gausMap[(y*mapW)+x] = (1/Math.pow(2*Math.PI*Math.pow(sigma,2),0.5)) * Math.exp(-1*(Math.pow(rad,2)/(2*Math.pow(sigma,2) ) ));
}

//-------output image of kernel-----------------
/*
float[] fGausMap = new float[mapW*mapW];
for(int i=0; i<mapW*mapW; i++)
	fGausMap[i] = (float) (gausMap[i]);
ImageStack kernel = new ImageStack(mapW, mapW);
	kernel.addSlice("gausMap", fGausMap);
new ImagePlus("Gaussian Kernel", kernel).show();
*/

// --------------------- Get ROI Mask and Outline---------------------------------

Roi roi = imp.getRoi();
if (roi!=null && !roi.isArea()) roi = null;
ImageProcessor ip = imp.getProcessor();
ImageProcessor mask = roi.getMask();

Rectangle r = roi.getBounds();

int w = ip.getWidth();
int h = ip.getHeight();

float[] pixels = new float[w*h];
double[] pixels2 = new double[r.width*r.height];
float[] counts = new float[w*h];
double[] counts2 = new double[r.width*r.height];

for (int y=0; y<r.height; y++)
for (int x=0; x<r.width; x++)
if(mask.getPixel(x,y) != 0){
	pixels[((y+r.y)*w) + x+r.x] = ip.getPixelValue(x+r.x, y+r.y);
	counts[((y+r.y)*w) + x+r.x] = 1;
}

int loc = 0;

for(int y=0; y<r.height; y++)
for(int x=0; x<r.width; x++)
if(mask.getPixel(x,y) != 0)
for(int y2=y-cutoff; y2<=y+cutoff; y2++)
for(int x2=x-cutoff; x2<=x+cutoff; x2++){
	if(mask.getPixel(x2,y2) != 0){
		loc = ((y2+cutoff-y)*mapW) + (x2+cutoff-x);
		pixels2[(y*r.width)+x] = pixels2[(y*r.width)+x] + (pixels[((y2+r.y)*w)+(x2+r.x)] * gausMap[loc]);
		counts2[(y*r.width)+x] = counts2[(y*r.width)+x] + gausMap[loc];
	}
	IJ.showProgress((float) y/r.height);
}

IJ.showProgress((float) 1);


for(int y=0; y<r.height; y++)
for(int x=0; x<r.width; x++)
if(mask.getPixel(x,y) != 0)
	 ip.setf(x+r.x, y+r.y, (float) (pixels2[(y*r.width)+x]/counts2[(y*r.width)+x]) );

imp.updateAndDraw();

}
}
