
/*__________________________________________________________________

	Title: QCPA Analysis
	''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Authors: Jolyon Troscianko & Cedric van den Berg
	Date: 18/3/2018
............................................................................................................

This script performs the QCPA adjacency analysis on a clustered image.

Requirements:

A cluster ID image, where pixel value integers refer to each cluster (starting
from cluster 1). Also required is a results table containing information about
each cluster and mean values of each receptor class (must end with "_mean").
There must be one receptor class for each receptor, plus a luminance channel.

e.g. the Results window for a trichromat should have column labels like:

label/clusterID	lw_mean	mw_mean	sw_mean	lum_mean

Where the row number refers to the clusterID (and must match the respective
pixel values in the cluster image).

If an ROI is selected, the analysis is only performed in the ROI.
____________________________________________________________________
*/


import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;


import ij.measure.ResultsTable;
import ij.util.Tools;

import ij.text.TextWindow;
import ij.plugin.filter.Analyzer;
import ij.WindowManager;

import ij.plugin.frame.RoiManager;
import ij.gui.Roi; 


public class QCPA_Analysis implements PlugIn {
public void run(String arg) {


IJ.showStatus("Adjacency Analysis");
IJ.log("_____________________________________________");
IJ.log("------------------Adjacency Analysis------------------");


ImagePlus imp = IJ.getImage();
Roi roi = imp.getRoi();

ImageProcessor ip = imp.getProcessor();
int w = ip.getWidth();
int h = ip.getHeight();

//float[] clusterLookups = new float[w*h];
int[] clusterLookups = new int[w*h];
for (int i=0; i<w*h; i++)
	clusterLookups[i] = -1; // -1 is the number used in the adjacency code to ignore clusters, so this saves me from changing that code



if( roi == null ){ //-----------------No selection - Measuring whole image-----------------
	IJ.log("Measuring whole image");

	for (int y=0; y<h; y++)
	for (int x=0; x<w; x++)
		clusterLookups[(y*w) + x] = (int) (ip.getPixelValue(x, y)-1);
} else { //-----------------Measuring ROI-----------------

	ImageProcessor mask = roi.getMask();
	Rectangle r = roi.getBounds();

	if(mask == null){
		IJ.log("Measuring rectangular selection");
		for (int y=r.y; y<r.y+r.height; y++)
		for (int x=r.x; x<r.x+r.width; x++)
			clusterLookups[(y*w) + x] = (int) (ip.getPixelValue(x, y)-1);
	} else {
		IJ.log("Measuring selected area");
		for (int y=r.y; y<r.y+r.height; y++)
		for (int x=r.x; x<r.x+r.width; x++)
		if(mask.getPixel(x-r.x,y-r.y) != 0)
			clusterLookups[(y*w) + x] = (int) (ip.getPixelValue(x, y)-1);
	}

}

/*
// Create output image for debugging
float[] fClusters = new float[w*h];
for (int i=0; i<w*h; i++)
	fClusters[i] = (float) (clusterLookups[i]);

ImageStack zones = new ImageStack(w, h);
	zones.addSlice("zones", fClusters);
new ImagePlus("Zones", zones).show();
*/

//-----------------------------Initialise Results Table----------------------------

String[] titles = WindowManager.getNonImageTitles();
int cResultsOpen = 0;
int sResultsOpen = 0;
int rResultsOpen = 0;

for(int i=0; i<titles.length; i++){
	if(titles[i] == "Cluster Results")
		cResultsOpen ++;
	if(titles[i] == "Summary Results")
		sResultsOpen ++;
	if(titles[i] == "ROI Cluster Results")
		rResultsOpen ++;
}

if(cResultsOpen > 0)
	IJ.renameResults("Cluster Results", "Results");

ResultsTable rt = Analyzer.getResultsTable();

if(rt == null) {
	IJ.log("Error - no cluster results window found");
	IJ.log("The clusters should be in a window called");
	IJ.log("'Cluster Results' or 'Results'. There must");
	IJ.log("be the same number of clusters in the");
	IJ.log("image and results window.");
	return;
}



//-----------------------Get values from Results Table----------------------
// First column should be group IDs, then alternating mean and SD for each dimension

String[] columns = rt.getHeadings();


int nDims = 0;
for(int i=0; i<columns.length; i++){
	String[] header = columns[i].split("_");
	if(header.length > 1)
		if("mean".equals(header[header.length-1]))
			nDims ++;
}

if(nDims == 0) {
	IJ.log("Error - no cluster mean values found");
	IJ.log("The clusters should be in a window called");
	IJ.log("'Cluster Results' or 'Results'. There must");
	IJ.log("be the same number of clusters in the");
	IJ.log("image and results window.");
	return;
}

IJ.log("Channels found (incl Lum): " + nDims);


double[] tempX = rt.getColumnAsDoubles(1);
int nClusters = tempX.length;
IJ.log("Clusters found in results table: " + nClusters);


String[] sliceNames = new String[nDims];
double[] cMeans = new double[nDims * nClusters];
int cChannel = 0; // save current channel

for(int i=0; i<columns.length; i++){
	String[] header = columns[i].split("_");
	if(header.length > 1)
		if("mean".equals(header[header.length-1])){
			sliceNames[cChannel] = header[0];
			tempX = rt.getColumnAsDoubles(i);
			for(int j=0; j<tempX.length; j++)
				cMeans[(cChannel*nClusters)+j] = tempX[j];
			cChannel ++;
		}
}

IJ.renameResults("Results", "Cluster Results");



//--------------------------------Get Settings-----------------------------

double[] weber = new double[nDims]; //

if(nDims == 4){
	//weber[0] = 0.05; // LW HUMAN scaled to 0.05 for LWS from Hofer et al 2005, assuming a 2:1 L:M and 5.72% S cones
	//weber[1] = 0.0707106781; // MW
	//weber[2] = 0.1657433633; // SW
	//weber[3] = 0.1; // 


	weber[0] = 0.05; // LW triggerfish
	weber[1] = 0.05; // MW
	weber[2] = 0.07; // SW
	weber[3] = 0.05; // 
} else {
for(int i = 0; i<nDims; i++)
	weber[i] = 0.05;

}

String imTitle = "ImageID";
double lumWeber = 0.05;


	GenericDialog gd = new GenericDialog("Adjacency Analysis");

		gd.addStringField("Image ID:", imTitle, 20);
		gd.addMessage(" ");
		gd.addMessage("Weber Fractions");

		for(int i=0; i<nDims-1; i++)
			//gd.addNumericField("Weber_" + sliceNames[i], weber[i], 3);	
			gd.addNumericField("Weber_" + (i+1) + " " + sliceNames[i], weber[i], 3);	
		
		gd.addNumericField("Luminance_Weber_Fraction", lumWeber, 3);
		gd.addCheckbox("Horizontal and vertical adjacency results", false);
		gd.addCheckbox("Output adjacency matrix", false);

	gd.showDialog();
	if (gd.wasCanceled())
		return;

imTitle = gd.getNextString();
for(int i=0; i<nDims-1; i++)
	weber[i] = gd.getNextNumber();
lumWeber = gd.getNextNumber();
boolean outputHVdata = gd.getNextBoolean();
boolean outMatrix = gd.getNextBoolean();

IJ.log("Label: " + imTitle);
for(int i=0; i<nDims-1; i++)
	IJ.log("weber "+ sliceNames[i] + ": " + weber[i] );
IJ.log("lumWeber: " + lumWeber);


double[] dfis = new double[nDims];
double JND = 0.0;


	if(rResultsOpen > 0)
		IJ.renameResults("ROI Cluster Results", "Results");

	ResultsTable ct = Analyzer.getResultsTable();

	if (ct == null) {
		ct = new ResultsTable();
		Analyzer.setResultsTable(ct);
	}



//---------------------------------------------------ADJACENCY ANALYSIS------------------------------------------

// Calculate proportional coverage of each cluster
double[] prCovers = new double[nClusters];
int npx = 0; // total number of pixels in ROI
int[] area = new int[nClusters];

for(int i=0; i<nClusters; i++)
for(int j=0; j<w*h; j++)
	if(clusterLookups[j] == i){
		prCovers[i] ++;
		area[i] ++;
		npx ++;
	}

int nClustersROI = 0;
for(int i=0; i<nClusters; i++)
	if(prCovers[i] > 0)
		nClustersROI ++;

IJ.log("Clusters found in ROI: " + nClustersROI );

for(int i=0; i<nClusters; i++){
	prCovers[i] = prCovers[i]/npx;
	//IJ.log("Cluster: " + (i+1) + " cover: " + prCovers[i]);
}

	// Calculate horizontal adjacency numbers
	int[] hAdjs = new int[nClusters*nClusters];
	double hOffCount = 0.0;
	double hDiaCount = 0.0;
	int hRunCount = 0; // values for calculating patch size - this counts the values within each run
	int hRunSum = 0; // this sums the total number

	for(int y=0; y<h; y++){
	for(int x=0; x<w-1; x++)
	if(clusterLookups[(y*w)+x] != -1 && clusterLookups[(y*w)+x+1] != -1){ //
		if(clusterLookups[(y*w)+x] != clusterLookups[(y*w)+x+1]){ // intersection of two clusters
			// fill both sides of the matrix
			hAdjs[ (clusterLookups[(y*w)+x+1] * nClusters) + clusterLookups[(y*w)+x]  ] ++;
			hAdjs[ (clusterLookups[(y*w)+x] * nClusters) + clusterLookups[(y*w)+x+1]  ] ++;
			hOffCount ++;
			hRunCount += hRunSum;
			hRunSum = 0;
			
		} else { // intersection of the same clusters
			hAdjs[(clusterLookups[(y*w)+x] * nClusters) + clusterLookups[(y*w)+x] ] ++;
			hDiaCount ++;
			hRunSum++;
		}
	} // x fi
	hRunCount += hRunSum;
	hRunSum = 0; // restart on new row
	} //y


	// Calculate vertical adjacency numbers
	int[] vAdjs = new int[nClusters*nClusters];
	double vOffCount = 0.0;
	double vDiaCount = 0.0;
	int vRunCount = 0; // values for calculating patch size - this counts the values within each run
	int vRunSum = 0; // this sums the total number

	for(int y=0; y<h-1; y++){
	for(int x=0; x<w; x++)
	if(clusterLookups[(y*w)+x] != -1 && clusterLookups[((y+1)*w)+x] != -1){ 
		if(clusterLookups[(y*w)+x] != clusterLookups[((y+1)*w)+x]){ // vertical intersection of two clusters
			// fill both sides of the matrix
			vAdjs[ (clusterLookups[((y+1)*w)+x] * nClusters) + clusterLookups[(y*w)+x]  ] ++;
			vAdjs[ (clusterLookups[(y*w)+x] * nClusters) + clusterLookups[((y+1)*w)+x]  ] ++;
			vOffCount ++;
			vRunCount += vRunSum;
			vRunSum = 0;
		} else {
			vAdjs[(clusterLookups[(y*w)+x] * nClusters) + clusterLookups[(y*w)+x] ] ++;
			vDiaCount ++;
			vRunSum++;
		}
	} // x fi
	vRunCount += vRunSum;
	vRunSum = 0; // restart on new column
	} //y

//------------count non-zero transition types-------------
	int nzOffCount = 0; // non-zero transition type count
	int hnzOffCount = 0;
	int vnzOffCount = 0;

	for(int i=0; i<nClusters; i++)
	for(int j=i+1; j<nClusters; j++){
		if(hAdjs[(i*nClusters)+j] > 0)
			hnzOffCount ++;
	
		if(vAdjs[(i*nClusters)+j] > 0)
			vnzOffCount ++;
	
		if(hAdjs[(i*nClusters)+j] > 0 || vAdjs[(i*nClusters)+j] > 0)
			nzOffCount ++;
	}

	IJ.log("n (non-zero off-diagonal transition types) = " + nzOffCount);

	//----------------Combine horizontal and vertical into a single matrix-----------------------
	
	int[] Adjs = new int[nClusters*nClusters];
	for(int i=0; i<nClusters*nClusters; i++)
		Adjs[i] = vAdjs[i] + hAdjs[i];

	// variables for calculating indices
	double scSum = 0.0;
	double stSum = 0.0;
	double HcSum = 0.0;
	double HtSum = 0.0;

	double hscSum = 0.0;
	double hstSum = 0.0;
	double vscSum = 0.0;
	double vstSum = 0.0;

	double hHcSum = 0.0;
	double hHtSum = 0.0;
	double vHcSum = 0.0;
	double vHtSum = 0.0;
	double tVal = 0.0;
	double pVal = 0.0;

	//--------------Chomraticity variables--------------

	double Dtemp = 0.0;
	String DmaxChannel = "";
	double[] Dmaxs = new double[nClusters];
	double[] Lums = new double[nClusters];
	double rnlX = 0.0;
	double rnlY = 0.0;
	double rnlZ = 0.0;
	double A = 0.0;
	double a = 0.0;
	double b = 0.0;
	double c = 0.0;
	double[] RNLChr = new double[nClusters];
	double c0 = 0.0;
	double c1 = 0.0;
	double c2 = 0.0;
	double c3 = 0.0;

	for(int k=0; k<nClusters; k++)
	if(area[k] > 0){
			ct.incrementCounter();
			ct.addValue("Image", imTitle);
			ct.addValue("ClusterID", (k+1));
			ct.addValue("Area", area[k]);
			ct.addValue("Coverage (pc)", prCovers[k]*100.0);
			for(int j=0; j<nDims; j++)
				ct.addValue(sliceNames[j] + "_mean", cMeans[(j*nClusters)+k]);
				
			//--------------Calculate RNL Chromaticity----------------------

			RNLChr[k] = 0.0;

			if(nDims >= 3){ // DICHROMAT
				c0 = cMeans[k] ;
				c1 = cMeans[(1*nClusters)+k] ;
				rnlX =  Math.pow(1/(Math.pow(weber[1],2)+Math.pow(weber[0], 2)), 0.5)*(Math.log(c0)-Math.log(c1));	
				RNLChr[k] += (rnlX * rnlX);
				ct.addValue("RNL X", rnlX);
			}
			if(nDims >= 4){ // TRICHROMAT
				c2 = cMeans[(2*nClusters)+k] ;
				rnlY = Math.pow((Math.pow(weber[1] , 2)+Math.pow(weber[0] , 2))/(Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2)+Math.pow(weber[1]*weber[0],2)) , 0.5)*(Math.log(c2)-(Math.log(c0)*(Math.pow(weber[1], 2)/(Math.pow(weber[1], 2)+Math.pow(weber[0],2)))+Math.log(c1)*(Math.pow(weber[0],2)/(Math.pow(weber[1], 2)+Math.pow(weber[0], 2)))))      ;
				RNLChr[k] += (rnlY * rnlY);
				ct.addValue("RNL Y", rnlY);
			}
			if(nDims >= 5){ // TETRACHROMAT
				c3 = cMeans[(3*nClusters)+k] ; 
				A = Math.pow((Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2))/(Math.pow(weber[2]*weber[1]*weber[0], 2)+Math.pow(weber[3]*weber[1]*weber[0], 2)+Math.pow(weber[3]*weber[2]*weber[0], 2)+Math.pow(weber[3]*weber[2]*weber[1], 2)),0.5);
				a = Math.pow(weber[2]*weber[1], 2)/(Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2));
				b = Math.pow(weber[2]*weber[0], 2)/(Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2));
				c = Math.pow(weber[1]*weber[0], 2)/(Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2));
				rnlZ =      A*(Math.log(c3)-(a*Math.log(c0)+b*Math.log(c1)+c*Math.log(c2)))     ;
				RNLChr[k] += (rnlZ * rnlZ);
				ct.addValue("RNL Z", rnlZ);
			}

			RNLChr[k] = Math.pow(RNLChr[k], 0.5);
			ct.addValue("RNL Chromaticity", RNLChr[k]);
			

			//---------------Calculate Chromaticity-----------------
			
			Dtemp = 0.0;
			Dmaxs[k] = 0; // these oponent chanels are on a -1 to 1 range, where zero is grey
			for(int j=0; j<nDims-1; j++)	
			for(int jj=j+1; jj<nDims-1; jj++){
				Dtemp = Math.abs(   ( cMeans[(j*nClusters)+k] - cMeans[(jj*nClusters)+k]  ) / (  cMeans[(j*nClusters)+k] + cMeans[(jj*nClusters)+k]  )  );
				//Dtemp = Math.abs(   (means[(nPoints*j)+i] - means[(nPoints*jj)+i]) / (means[(nPoints*j)+i] + means[(nPoints*jj)+i])  );
				if(Dtemp > Dmaxs[k]){
					Dmaxs[k] = Dtemp;
					DmaxChannel = sliceNames[j] +":"+sliceNames[jj];
				}
				//Daves[k] += Dtemp;
			}

			//if(outMatrix == true){
				ct.addValue("Dmax Chromaticity", Dmaxs[k]);
				ct.addValue("Dmax channel", DmaxChannel);
				//ct.addValue("Dave Chromaticity", Daves[k]);
			//}

			//Lums[k] = means[(nPoints*(nDims-1))+i];
			Lums[k] = cMeans[(nClusters*(nDims-1))+k]; // luminance channel must be last channel

			//-------------Matrix calculations------------------
			for(int l=0; l<nClusters; l++)
			if(area[l] > 0){
				if(l == k){
					pVal = Adjs[(k*nClusters)+l] / (vDiaCount+hDiaCount);
					if(outMatrix == true)
						ct.addValue("Adj"+(l+1), Adjs[(k*nClusters)+l]);
					scSum = scSum + Math.pow(pVal,2);
					if(pVal > 0)
						HcSum = HcSum + ( pVal * Math.log(pVal));
						
				} else {
					tVal = Adjs[(k*nClusters)+l] / (vOffCount+hOffCount);
					if(outMatrix == true)
						ct.addValue("Adj"+(l+1), Adjs[(k*nClusters)+l]);
					if(l<k){
						stSum = stSum + Math.pow(tVal,2);
						if(tVal > 0)
							HtSum = HtSum + ( tVal * Math.log(tVal));
							
					}
				}
			}
			
			//-------------Horizontal matrix calculations------------------
			for(int l=0; l<nClusters; l++)
			if(area[l] > 0){
				if(l == k){
					pVal = hAdjs[(k*nClusters)+l] / hDiaCount;
					if(outMatrix == true)
					if(outputHVdata == true)
						ct.addValue("hAdj"+(l+1), hAdjs[(k*nClusters)+l]);
					hscSum = hscSum + Math.pow(pVal,2);
					if(pVal > 0)
						hHcSum = hHcSum + ( pVal * Math.log(pVal));
						
				} else {
					tVal = hAdjs[(k*nClusters)+l] / hOffCount;
					if(outMatrix == true)
					if(outputHVdata == true)
						ct.addValue("hAdj"+(l+1), hAdjs[(k*nClusters)+l]);
					if(l<k){
						hstSum = hstSum + Math.pow(tVal,2);
						if(tVal > 0)
							hHtSum = hHtSum + ( tVal * Math.log(tVal));
							
					}
				}
			}

			//-------------Vertical matrix calculations------------------
			for(int l=0; l<nClusters; l++)
			if(area[l] > 0){
				if(l == k){
					pVal = vAdjs[(k*nClusters)+l] / vDiaCount;
					if(outMatrix == true)
					if(outputHVdata == true)
						ct.addValue("vAdj"+(l+1), vAdjs[(k*nClusters)+l]);
					vscSum = vscSum + Math.pow(pVal,2);
					if(pVal > 0)
						vHcSum = vHcSum + ( pVal * Math.log(pVal));
						
				} else {
					tVal = vAdjs[(k*nClusters)+l] / vOffCount;
					if(outMatrix == true)
					if(outputHVdata == true)
						ct.addValue("vAdj"+(l+1), vAdjs[(k*nClusters)+l]);
					if(l<k){
						vstSum = vstSum + Math.pow(tVal,2);
						if(tVal > 0)
							vHtSum = vHtSum + ( tVal * Math.log(tVal));
							
					}
				}
			}

			
	} // k


	ct.showRowNumbers(false);
	ct.show("Results");
	IJ.renameResults("Results", "ROI Cluster Results");

	if(sResultsOpen > 0)
		IJ.renameResults("Summary Results", "Results");
	
	ResultsTable st = Analyzer.getResultsTable();
	if (st == null) {
		st = new ResultsTable();
		Analyzer.setResultsTable(st);
	}

	st.incrementCounter();
	st.addValue("Image", imTitle);



	//---------------------------------------CoV results (on-diagonals, though I'm using actual cluster count sizes)--------------------------------------------

	double meanDmax = 0.0;
	double meanRNLChr = 0.0;
	double meanLum = 0.0;

	double hmeanDmax = 0.0;
	double hmeanRNLChr = 0.0;
	double hmeanLum = 0.0;

	double vmeanDmax = 0.0;
	double vmeanRNLChr = 0.0;
	double vmeanLum = 0.0;

	for(int k=0; k<nClusters; k++)
	if(area[k] > 0){
		meanDmax += Dmaxs[k] * prCovers[k]; // chromaticiy multiplied by proportion of image coverage
		meanRNLChr += RNLChr[k] * prCovers[k];
		meanLum += Lums[k] * prCovers[k];

		hmeanDmax += Dmaxs[k] * (hAdjs[(k*nClusters)+k] / hDiaCount); // chromaticiy multiplied by matrix diagonal counts
		hmeanRNLChr += RNLChr[k] * (hAdjs[(k*nClusters)+k] / hDiaCount);
		hmeanLum += Lums[k] * (hAdjs[(k*nClusters)+k] / hDiaCount);

		vmeanDmax += Dmaxs[k] * (vAdjs[(k*nClusters)+k] / vDiaCount); // chromaticiy multiplied by matrix diagonal counts
		vmeanRNLChr += RNLChr[k] * (vAdjs[(k*nClusters)+k] / vDiaCount);
		vmeanLum += Lums[k] * (vAdjs[(k*nClusters)+k] / vDiaCount);
	}

	// sums of squares
	double DmaxSS = 0.0;
	double RNLChrSS = 0.0;
	double LumSS = 0.0;

	double hDmaxSS = 0.0;
	double hRNLChrSS = 0.0;
	double hLumSS = 0.0;

	double vDmaxSS = 0.0;
	double vRNLChrSS = 0.0;
	double vLumSS = 0.0;

	for(int k=0; k<nClusters; k++)
	if(area[k] > 0){
		DmaxSS += prCovers[k] * Math.pow(Dmaxs[k] - meanDmax,2);
		RNLChrSS += prCovers[k] * Math.pow(RNLChr[k] - meanRNLChr,2);
		LumSS += prCovers[k] * Math.pow(Lums[k] - meanLum,2);

		hDmaxSS += (hAdjs[(k*nClusters)+k] / hDiaCount) * Math.pow(Dmaxs[k] - hmeanDmax,2);
		hRNLChrSS += (hAdjs[(k*nClusters)+k] / hDiaCount) * Math.pow(RNLChr[k] - hmeanRNLChr,2);
		hLumSS += (hAdjs[(k*nClusters)+k] / hDiaCount) * Math.pow(Lums[k] - hmeanLum,2);

		vDmaxSS += (vAdjs[(k*nClusters)+k] / vDiaCount) * Math.pow(Dmaxs[k] - vmeanDmax,2);
		vRNLChrSS += (vAdjs[(k*nClusters)+k] / vDiaCount) * Math.pow(RNLChr[k] - vmeanRNLChr,2);
		vLumSS += (vAdjs[(k*nClusters)+k] / vDiaCount) * Math.pow(Lums[k] - vmeanLum,2);
	}

	// convert to standard deviation and then Coefficient of variation

	DmaxSS =  Math.pow( (DmaxSS * nClustersROI) / (nClustersROI-1), 0.5);
	RNLChrSS =  Math.pow( (RNLChrSS * nClustersROI) / (nClustersROI-1), 0.5);
	LumSS =  Math.pow( (LumSS * nClustersROI) / (nClustersROI-1), 0.5);

	hDmaxSS =  Math.pow( (hDmaxSS * nClustersROI) / (nClustersROI-1), 0.5);
	hRNLChrSS =  Math.pow( (hRNLChrSS * nClustersROI) / (nClustersROI-1), 0.5);
	hLumSS =  Math.pow( (hLumSS * nClustersROI) / (nClustersROI-1), 0.5);

	vDmaxSS =  Math.pow( (vDmaxSS * nClustersROI) / (nClustersROI-1), 0.5);
	vRNLChrSS =  Math.pow( (vRNLChrSS * nClustersROI) / (nClustersROI-1), 0.5);
	vLumSS =  Math.pow( (vLumSS * nClustersROI) / (nClustersROI-1), 0.5);



	//---------------------------------------BOUNDARY STRENGTH CoV for transitions (off-diagonals of matrix)-----------------------------------

	double meanDmaxT = 0.0;
	double meanRNLChrT = 0.0;
	double meanLumT = 0.0;
	double meanJNDT = 0.0;

	double meanMichLumT = 0.0;

	double hmeanDmaxT = 0.0;
	double hmeanRNLChrT = 0.0;
	double hmeanLumT = 0.0;
	double hmeanJNDT = 0.0;

	double hmeanMichLumT = 0.0;

	double vmeanDmaxT = 0.0;
	double vmeanRNLChrT = 0.0;
	double vmeanLumT = 0.0;
	double vmeanJNDT = 0.0;

	double vmeanMichLumT = 0.0;

	int nOffDiagonals =0;
	int nOffDiagonalZeros = 0;
	int hOffDiagonalZeros = 0;
	int vOffDiagonalZeros = 0;

	// Euclidean chromaticity or JND differences multiplied by proportion of transitions
	for(int k=0; k<nClusters; k++)
	if(area[k] > 0)
	for(int l=k+1; l<nClusters; l++)
	if(area[l]>0){
		if(Adjs[(k*nClusters)+l] == 0)
			nOffDiagonalZeros ++;
		else {
		//--------------Average boundaries------------
		tVal = Adjs[(k*nClusters)+l] / (vOffCount+hOffCount); // proportion

		meanDmaxT += Math.abs((Dmaxs[k]-Dmaxs[l])/(Dmaxs[k]+Dmaxs[l])) * tVal; 
		meanRNLChrT += Math.abs((RNLChr[k]-RNLChr[l])/(RNLChr[k]+RNLChr[l])) * tVal; 

		// ------------------Luminance (Siddiqi) JNDs----------------:
		meanLumT += Math.abs((Math.log( Lums[k] )-Math.log( Lums[l] ))/lumWeber) * tVal;

		// ------------------Michellson Contrasts----------------:
		meanMichLumT += Math.abs( (Lums[k] - Lums[l]) / (Lums[k] + Lums[l]) )    * tVal;

		// -----------------Chromatic JNDs----------------------
		for(int i=0; i<nDims-1; i++)
			dfis[i] = Math.log(cMeans[(i*nClusters)+k]/cMeans[(i*nClusters)+l]);
			//dfis[i] = Math.log(cMeans[(k*(nDims-1))+i]/cMeans[(l*(nDims-1))+i]);
		if(nDims == 3) // DICHROMATIC EQUATION
			meanJNDT += tVal * Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);
		else if(nDims == 4) // TRICHROMATIC EQUATION
			meanJNDT += tVal * Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);
		else if(nDims == 5) // TETRACHROMATIC EQUATION
			meanJNDT += tVal * Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);
		}

		if(hAdjs[(k*nClusters)+l] == 0)
			hOffDiagonalZeros ++;
		else {
		//-------------Horizontal boundaries--------------
		tVal = hAdjs[(k*nClusters)+l] / hOffCount;
		hmeanDmaxT += Math.abs((Dmaxs[k]-Dmaxs[l])/(Dmaxs[k]+Dmaxs[l])) * tVal; 
		hmeanRNLChrT += Math.abs((RNLChr[k]-RNLChr[l])/(RNLChr[k]+RNLChr[l])) * tVal; 

		// ------------------Horizontal Luminance (Siddiqi) JNDs----------------:
		hmeanLumT += Math.abs((Math.log( Lums[k] )-Math.log( Lums[l] ))/lumWeber) * tVal;
		// ------------------Michellson Contrasts----------------:
		hmeanMichLumT += Math.abs( (Lums[k] - Lums[l]) / (Lums[k] + Lums[l]) )    * tVal;
		// -----------------Horizontal Chromatic JNDs----------------------
		if(nDims == 3) // DICHROMATIC EQUATION
			hmeanJNDT += tVal * Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);
		else if(nDims == 4) // TRICHROMATIC EQUATION
			hmeanJNDT += tVal * Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);
		else if(nDims == 5) // TETRACHROMATIC EQUATION
			hmeanJNDT += tVal * Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);
		}

		if(vAdjs[(k*nClusters)+l] == 0)
			vOffDiagonalZeros ++;
		else {
		//-------------Vertical boundaries--------------
		tVal = vAdjs[(k*nClusters)+l] / vOffCount;
		vmeanDmaxT += Math.abs((Dmaxs[k]-Dmaxs[l])/(Dmaxs[k]+Dmaxs[l])) * tVal; 
		vmeanRNLChrT += Math.abs((RNLChr[k]-RNLChr[l])/(RNLChr[k]+RNLChr[l])) * tVal; 

		// ------------------Vertical Luminance (Siddiqi) JNDs----------------:
		vmeanLumT += Math.abs((Math.log( Lums[k] )-Math.log( Lums[l] ))/lumWeber) * tVal;
		// ------------------Michellson Contrasts----------------:
		vmeanMichLumT += Math.abs( (Lums[k] - Lums[l]) / (Lums[k] + Lums[l]) )    * tVal;
		// -----------------Vertical Chromatic JNDs----------------------
		if(nDims == 3) // DICHROMATIC EQUATION
			vmeanJNDT += tVal * Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);
		else if(nDims == 4) // TRICHROMATIC EQUATION
			vmeanJNDT += tVal * Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);
		else if(nDims == 5) // TETRACHROMATIC EQUATION
			vmeanJNDT += tVal * Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);
		}

		nOffDiagonals ++;
		
	}



	if(nOffDiagonals != (nClustersROI*nClustersROI-nClustersROI)/2)
		IJ.log("Error - matrix doesn't add up");

	// sums of squares
	double DmaxSST = 0.0;
	double RNLChrSST = 0.0;
	double LumSST = 0.0;
	double JNDSST = 0.0;
	double michLumSST = 0.0;

	double hDmaxSST = 0.0;
	double hRNLChrSST = 0.0;
	double hLumSST = 0.0;
	double hJNDSST = 0.0;
	double hmichLumSST = 0.0;

	double vDmaxSST = 0.0;
	double vRNLChrSST = 0.0;
	double vLumSST = 0.0;
	double vJNDSST = 0.0;
	double vmichLumSST = 0.0;

	double colJND = 0.0;
	double lumJND = 0.0;

	double lumContr = 0.0; // weighted mean luminance RNL contrast
	double[] lumContrs = new double[nOffDiagonals];
	double colContr = 0.0; // weighted mean chromatic RNL contrast
	double[] colContrs = new double[nOffDiagonals];
	double[] ffs = new double[nOffDiagonals];
	double ffSum = 0.0;

	double hlumContr = 0.0; // weighted mean luminance RNL contrast
	double hcolContr = 0.0; // weighted mean chromatic RNL contrast
	double[] hffs = new double[nOffDiagonals];
	double hffSum = 0.0;

	double vlumContr = 0.0; // weighted mean luminance RNL contrast
	double vcolContr = 0.0; // weighted mean chromatic RNL contrast
	double[] vffs = new double[nOffDiagonals];
	double vffSum = 0.0;


	int tCount = 0;

	
	for(int k=0; k<nClusters; k++)
	if(area[k] > 0)
	for(int l=k+1; l<nClusters; l++)
	if(area[l]>0){
		//--------------Average boundaries------------
		tVal = Adjs[(k*nClusters)+l] / (vOffCount+hOffCount); // proportion
		RNLChrSST += tVal * Math.pow( Math.abs((RNLChr[k]-RNLChr[l])/(RNLChr[k]+RNLChr[l])) - meanRNLChrT	,2);

		//----------------Michelson Contrast--------------
		michLumSST += tVal * Math.pow(Math.abs( (Lums[k] - Lums[l]) / (Lums[k] + Lums[l]) ) - meanMichLumT	,2);

		// ------------------Luminance (Siddiqi) JNDs----------------:
		lumJND = Math.abs((Math.log( Lums[k] )-Math.log( Lums[l] ))/lumWeber);
		LumSST += tVal * Math.pow(		lumJND - meanLumT		,2);

		ffs[tCount] = ((Adjs[(k*nClusters)+k] / (vDiaCount+hDiaCount)) + (Adjs[(l*nClusters)+l] / (vDiaCount+hDiaCount)))/2;
		ffSum += ffs[tCount];
		lumContrs[tCount] = lumJND;
		lumContr += ffs[tCount] * lumJND;

		hffs[tCount] = ((hAdjs[(k*nClusters)+k] / hDiaCount) + (hAdjs[(l*nClusters)+l] / hDiaCount))/2;
		hffSum += hffs[tCount];
		hlumContr += hffs[tCount] * lumJND;

		vffs[tCount] = ((vAdjs[(k*nClusters)+k] / vDiaCount) + (vAdjs[(l*nClusters)+l] / vDiaCount))/2;
		vffSum += vffs[tCount];
		vlumContr += vffs[tCount] * lumJND;


		DmaxSST += tVal * Math.pow( (Math.abs((Dmaxs[k]-Dmaxs[l])/(Dmaxs[k]+Dmaxs[l])) - meanDmaxT ), 2);

	
		// -----------------Chromatic JNDs----------------------
		for(int i=0; i<nDims-1; i++)
			dfis[i] = Math.log(cMeans[(i*nClusters)+k]/cMeans[(i*nClusters)+l]);
			//dfis[i] = Math.log(cMeans[(k*(nDims-1))+i]/cMeans[(l*(nDims-1))+i]);
		if(nDims == 3) // DICHROMATIC EQUATION
			colJND = Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);
			//JNDSST += tVal * Math.pow(Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5) - meanJNDT, 2);
		else if(nDims == 4) // TRICHROMATIC EQUATION
			colJND = Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);
			//JNDSST += tVal * Math.pow(Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5) - meanJNDT, 2);
		else if(nDims == 5) // TETRACHROMATIC EQUATION
			colJND = Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);
			//JNDSST += tVal * Math.pow(Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5) - meanJNDT, 2);

		JNDSST += tVal * Math.pow(colJND - meanJNDT, 2);
		

		colContrs[tCount] = colJND;
		colContr += ffs[tCount] * colJND;

		hcolContr += hffs[tCount] * colJND;

		vcolContr += vffs[tCount] * colJND;

		tCount++;

		//-------------Horizontal boundaries--------------
		tVal = hAdjs[(k*nClusters)+l] / hOffCount;
		hDmaxSST += tVal * Math.pow( (Math.abs((Dmaxs[k]-Dmaxs[l])/(Dmaxs[k]+Dmaxs[l])) - hmeanDmaxT ), 2);
		hRNLChrSST += tVal * Math.pow( Math.abs((RNLChr[k]-RNLChr[l])/(RNLChr[k]+RNLChr[l])) - hmeanRNLChrT	,2);

		//----------------Michelson Contrast--------------
		hmichLumSST += tVal * Math.pow(Math.abs( (Lums[k] - Lums[l]) / (Lums[k] + Lums[l]) ) - hmeanMichLumT	,2);
		// ------------------Horizontal Luminance (Siddiqi) JNDs----------------:
		hLumSST += tVal * Math.pow(		Math.abs((Math.log( Lums[k] )-Math.log( Lums[l] ))/lumWeber) - hmeanLumT		,2);


		// -----------------Horizontal Chromatic JNDs----------------------
		hJNDSST += tVal * Math.pow(colJND - hmeanJNDT, 2);

		//-------------Vertical boundaries--------------
		tVal = vAdjs[(k*nClusters)+l] / vOffCount;
		vDmaxSST += tVal * Math.pow( (Math.abs((Dmaxs[k]-Dmaxs[l])/(Dmaxs[k]+Dmaxs[l])) - vmeanDmaxT ), 2);
		vRNLChrSST += tVal * Math.pow( Math.abs((RNLChr[k]-RNLChr[l])/(RNLChr[k]+RNLChr[l])) - vmeanRNLChrT	,2);

		//----------------Michelson Contrast--------------
		vmichLumSST += tVal * Math.pow(Math.abs( (Lums[k] - Lums[l]) / (Lums[k] + Lums[l]) ) - vmeanMichLumT	,2);
		// ------------------Vertical Luminance (Siddiqi) JNDs----------------:
		vLumSST += tVal * Math.pow(		Math.abs((Math.log( Lums[k] )-Math.log( Lums[l] ))/lumWeber) - vmeanLumT	,2);
		// -----------------Vertical Chromatic JNDs----------------------
		vJNDSST += tVal * Math.pow(colJND - vmeanJNDT, 2);

	}


	colContr = colContr/ffSum;
	lumContr = lumContr/ffSum; // ffSum seems to be 1.5

	hcolContr = hcolContr/hffSum;
	hlumContr = hlumContr/hffSum; // ffSum seems to be 1.5

	vcolContr = vcolContr/vffSum;
	vlumContr = vlumContr/vffSum; // ffSum seems to be 1.5

	double colContrSS = 0.0;
	double lumContrSS = 0.0;

	double hcolContrSS = 0.0;
	double hlumContrSS = 0.0;

	double vcolContrSS = 0.0;
	double vlumContrSS = 0.0;

	for(int k=0; k<nOffDiagonals; k++){
		colContrSS += ffs[k] * Math.pow(colContrs[k]-colContr, 2);
		lumContrSS += ffs[k] * Math.pow(lumContrs[k]-lumContr, 2);

		hcolContrSS += hffs[k] * Math.pow(colContrs[k]-hcolContr, 2);
		hlumContrSS += hffs[k] * Math.pow(lumContrs[k]-hlumContr, 2);

		vcolContrSS += vffs[k] * Math.pow(colContrs[k]-vcolContr, 2);
		vlumContrSS += vffs[k] * Math.pow(lumContrs[k]-vlumContr, 2);
	}

	colContrSS = Math.pow( (nClustersROI * colContrSS) / ((nClustersROI-1) * ffSum) , 0.5);
	lumContrSS = Math.pow( (nClustersROI * lumContrSS) / ((nClustersROI-1) * ffSum) , 0.5);

	hcolContrSS = Math.pow( (nClustersROI * hcolContrSS) / ((nClustersROI-1) * hffSum) , 0.5);
	hlumContrSS = Math.pow( (nClustersROI * hlumContrSS) / ((nClustersROI-1) * hffSum) , 0.5);

	vcolContrSS = Math.pow( (nClustersROI * vcolContrSS) / ((nClustersROI-1) * vffSum) , 0.5);
	vlumContrSS = Math.pow( (nClustersROI * vlumContrSS) / ((nClustersROI-1) * vffSum) , 0.5);


	// convert to standard deviation and then Coefficient of variation

	DmaxSST =  Math.pow(	(nzOffCount*DmaxSST)/(nzOffCount-1)	,0.5);
	RNLChrSST =  Math.pow(	(nzOffCount*RNLChrSST)/(nzOffCount-1)	,0.5);
	LumSST =  Math.pow(	(nzOffCount*LumSST)/(nzOffCount-1)	,0.5);
	JNDSST =  Math.pow(	(nzOffCount*JNDSST)/(nzOffCount-1)	,0.5);
	michLumSST =  Math.pow(	(nzOffCount*michLumSST)/(nzOffCount-1)	,0.5);

	hDmaxSST =  Math.pow(	(hnzOffCount*hDmaxSST)/(hnzOffCount-1)	,0.5);
	hRNLChrSST =  Math.pow(	(hnzOffCount*hRNLChrSST)/(hnzOffCount-1)	,0.5);
	hLumSST =  Math.pow(	(hnzOffCount*hLumSST)/(hnzOffCount-1)	,0.5);
	hJNDSST =  Math.pow(	(hnzOffCount*hJNDSST)/(hnzOffCount-1)	,0.5);
	hmichLumSST =  Math.pow(	(hnzOffCount*hmichLumSST)/(hnzOffCount-1)	,0.5);

	vDmaxSST =  Math.pow(	(vnzOffCount*vDmaxSST)/(vnzOffCount-1)	,0.5);
	vRNLChrSST =  Math.pow(	(vnzOffCount*vRNLChrSST)/(vnzOffCount-1)	,0.5);
	vLumSST =  Math.pow(	(vnzOffCount*vLumSST)/(vnzOffCount-1)	,0.5);
	vJNDSST =  Math.pow(	(vnzOffCount*vJNDSST)/(vnzOffCount-1)	,0.5);
	vmichLumSST =  Math.pow(	(vnzOffCount*vmichLumSST)/(vnzOffCount-1)	,0.5);


	//-------------------------------Output Adjacency Results (H&V matrices combined)--------------------------

	// Inverse Simpson index
	st.addValue("CAA:Sc", 1/scSum); // Simpson Proportion Diversity Sc

	double Jc = (1/scSum)/nClustersROI;
	
	st.addValue("CAA:Jc", Jc); // Relative Diversity

	st.addValue("CAA:St", 1/stSum); // Simpson Transition Diversity

	double Jt =  (1/stSum)/  (( (nClustersROI) * (nClustersROI - 1) ) / 2);

	if(nClustersROI > 1){
		st.addValue("CAA:Jt", Jt ); // Relative Diversity
	} else {
		IJ.log("Too few clusters to calculate all diversity indices");
		st.addValue("CAA:Jt", "NaN" ); // Relative Diversity
	}

	// Shannon index
	st.addValue("CAA:Hc", (-1*HcSum)); // Shannon Div Proportion Diversity Sc

	double hMax = Math.log(nClustersROI);
	double Qc = (-1*HcSum)/hMax;

	if(nClustersROI > 1){
		st.addValue("CAA:Qc", Qc); // Shannon Div Normalised Shannon Colour Diversity
	} else {
		st.addValue("CAA:Qc", "NaN"); // Shannon Div Normalised Shannon Colour Diversity
	}

	st.addValue("CAA:Ht", (-1*HtSum)); // Shannon Div Transition Diversity

	double Qt = (-1*HtSum)/Math.log(nzOffCount);


	if(nClustersROI > 1){
		st.addValue("CAA:Qt", Qt); // Shannon Div Normalised Shannon Transition Diversity	
	} else {
		st.addValue("CAA:Qt", "NaN"); // Shannon Div Normalised Shannon Transition Diversity	
	}

	st.addValue("CAA:Scpl", (Jt+Jc)/2  ); //   Simpson colour complexity  previous version was  (1-Jt)+(1-Jc)

	if(nClustersROI > 1){
		st.addValue("CAA:Qcpl", (  Qt + Qc  )/2  ); // Shannon Colour Complexity
	} else {
		st.addValue("CAA:Qcpl", "NaN"  ); //
	}

	st.addValue("CAA:C", 1-(vDiaCount+hDiaCount)/((vDiaCount+hDiaCount)+ (vOffCount+hOffCount)) ); // Pattern Complexity


	//-----------------------------------Output Horizontal and Vertical Matrices separately------------------------
	if(outputHVdata == true){
	
		// Inverse Simpson index
		st.addValue("CAA:Sc Hrz", 1/hscSum); //Horizontal Proportion Diversity Sc
		st.addValue("CAA:Sc Vrt", 1/vscSum); //Vertical Proportion Diversity Sc

		double hJc = (1/hscSum)/nClustersROI;
		double vJc = (1/vscSum)/nClustersROI;

		st.addValue("CAA:Jc Hrz", hJc); //Horizontal Relative Diversity
		st.addValue("CAA:Jc Vrt", vJc); //Vertical Relative Diversity

		st.addValue("CAA:St Hrz", 1/hstSum); //Horizontal Transition Diversity
		st.addValue("CAA:St Vrt", 1/vstSum); //Vertical Transition Diversity

		double hJt =  (1/hstSum)/  (( (nClustersROI) * (nClustersROI - 1) ) / 2);
		double vJt = (1/vstSum)/    (( (nClustersROI) * (nClustersROI - 1) ) / 2);

		if(nClustersROI > 1){
			st.addValue("CAA:Jt Hrz", hJt ); //Horizontal Relative Diversity
			st.addValue("CAA:Jt Vrt", vJt ); //Vertical Relative Diversity
		} else {
			st.addValue("CAA:Jt Hrz", "NaN" ); //Horizontal Relative Diversity
			st.addValue("CAA:Jt Vrt", "NaN" ); //Vertical Relative Diversity
		}

		// Shannon index
		st.addValue("CAA:Hc Hrz", (-1*hHcSum)); //Horizontal Proportion Diversity Sc
		st.addValue("CAA:Hc Vrt", (-1*vHcSum)); //Vertical Proportion Diversity Sc

		double hQt = (-1*hHtSum)/hMax;
		double vQt = (-1*vHtSum)/hMax;
		double hQc = (-1*hHcSum)/Math.log(hnzOffCount);
		double vQc = (-1*vHcSum)/Math.log(vnzOffCount);

		if(nClustersROI > 1){
			st.addValue("CAA:Qc Hrz", hQc); // Normalised Shannon Colour Diversity	
			st.addValue("CAA:Qc Vrt", vQc); // Normalised Shannon Colour Diversity
		} else {
			st.addValue("CAA:Qc Hrz", "NaN"); // Normalised Shannon Colour Diversity	
			st.addValue("CAA:Qc Vrt", "NaN"); // Normalised Shannon Colour Diversit
		}

		st.addValue("CAA:Ht Hrz", (-1*hHtSum)); //Horizontal Transition Diversity
		st.addValue("CAA:Ht Vrt", (-1*vHtSum)); //Vertical Transition Diversity

		if(nClustersROI > 1){
			st.addValue("CAA:Qt Hrz", hQt); // Normalised Shannon Transition Diversity
			st.addValue("CAA:Qt Vrt", vQt); // Normalised Shannon Transition Diversity	
		} else {
			st.addValue("CAA:Qt Hrz", "NaN"); // Normalised Shannon Transition Diversity
			st.addValue("CAA:Qt Vrt", "NaN"); // Normalised Shannon Transition Diversity	
		}

		st.addValue("CAA:Scpl Hrz", (hJt + hJc)/2  ); //
		st.addValue("CAA:Scpl Vrt", (vJt + vJc)/2 ); //

		if(nClustersROI > 1){
			st.addValue("CAA:Qcpl Hrz", (  hQt + hQc  )/2  ); //
			st.addValue("CAA:Qcpl Vrt", (  vQt + vQc  )/2  ); //
		} else {
			st.addValue("CAA:Qcpl Hrz", "NaN" ); //
			st.addValue("CAA:Qcpl Vrt", "NaN" ); //
		}

		st.addValue("CAA:C Hrz", 1-( hDiaCount/(hDiaCount+ hOffCount) )); //
		st.addValue("CAA:C Vrt", 1-(vDiaCount/(vDiaCount+ vOffCount) ) ); //


	}// outputHVdata

	double AveRunLength = (double) (hRunCount + vRunCount) / (double) ( hOffCount + vOffCount );
	double hAveRunLength = (double) (hRunCount) / (double) (hOffCount);
	double vAveRunLength = (double) (vRunCount) / (double) (vOffCount);

	st.addValue("CAA:PT", AveRunLength ); // Patch Size Pt (px)
	st.addValue("CAA:PT Hrz", hAveRunLength ); // Patch Size Pt (px)
	st.addValue("CAA:PT Vrt", vAveRunLength ); // Patch Size Pt (px)
	st.addValue("CAA:Asp", hAveRunLength /( hAveRunLength + vAveRunLength) ); // Aspect ratio



	//----------------------Visual Contrast Analysis---------------------------


	st.addValue("VCA:ML", meanLum); // overall average luminance (not very interesting)
	st.addValue("VCA:sL", LumSS); //Luminance patch weighted SD
	st.addValue("VCA:CVL", LumSS/meanLum); //CoV

	st.addValue("VCA:MDmax", meanDmax); // Dmax patch weighted mean
	st.addValue("VCA:sDmax", DmaxSS); //Dmax patch weighted SD
	st.addValue("VCA:CVDmax", DmaxSS/meanDmax); // average CoV

	st.addValue("VCA:MSsat", meanRNLChr); // overall average chromaticity weighted by image coverage
	st.addValue("VCA:sSsat", RNLChrSS); //RNL Chr patch weighted SD
	st.addValue("VCA:CVSsat", RNLChrSS/meanRNLChr); // CoV

	st.addValue("VCA:MSL", lumContr); // weighted mean luminance RNL contrast
	st.addValue("VCA:sSL", lumContrSS); // weighted SD luminance RNL contrast
	st.addValue("VCA:CVSL", lumContrSS/lumContr); // CoV

	st.addValue("VCA:MS", colContr); // weighted mean chromatic RNL contrast
	st.addValue("VCA:sS", colContrSS); // weighted SD chromatic RNL contrast
	st.addValue("VCA:CVS", colContrSS/colContr); // CoV

	if(outputHVdata == true){

		st.addValue("VCA:ML Hrz", hmeanLum); // overall average luminance (not very interesting)
		st.addValue("VCA:sL Hrz", hLumSS); //Luminance patch weighted SD
		st.addValue("VCA:CVL Hrz", hLumSS/hmeanLum); //CoV

		st.addValue("VCA:ML Vrt", vmeanLum); // overall average luminance (not very interesting)
		st.addValue("VCA:sL Vrt", vLumSS); //Luminance patch weighted SD
		st.addValue("VCA:CVL Vrt", vLumSS/vmeanLum); //CoV

		st.addValue("VCA:MDmax Hrz", hmeanDmax); // Dmax patch weighted mean
		st.addValue("VCA:sDmax Hrz", hDmaxSS); //Dmax patch weighted SD
		st.addValue("VCA:CVDmax Hrz", hDmaxSS/hmeanDmax); // average CoV

		st.addValue("VCA:MDmax Vrt", vmeanDmax); // Dmax patch weighted mean
		st.addValue("VCA:sDmax Vrt", vDmaxSS); //Dmax patch weighted SD
		st.addValue("VCA:CVDmax Vrt", vDmaxSS/vmeanDmax); // average CoV

		st.addValue("VCA:MSsat Hrz", hmeanRNLChr); // overall average chromaticity weighted by image coverage
		st.addValue("VCA:sSsat Hrz", hRNLChrSS); //RNL Chr patch weighted SD
		st.addValue("VCA:CVSsat Hrz", hRNLChrSS/hmeanRNLChr); // CoV

		st.addValue("VCA:MSsat Vrt", vmeanRNLChr); // overall average chromaticity weighted by image coverage
		st.addValue("VCA:sSsat Vrt", vRNLChrSS); //RNL Chr patch weighted SD
		st.addValue("VCA:CVSsat Vrt", vRNLChrSS/vmeanRNLChr); // CoV

		st.addValue("VCA:MSL Hrz", hlumContr); // weighted mean luminance RNL contrast
		st.addValue("VCA:sSL Hrz", hlumContrSS); // weighted SD luminance RNL contrast
		st.addValue("VCA:CVSL Hrz", hlumContrSS/hlumContr); // CoV

		st.addValue("VCA:MSL Vrt", vlumContr); // weighted mean luminance RNL contrast
		st.addValue("VCA:sSL Vrt", vlumContrSS); // weighted SD luminance RNL contrast
		st.addValue("VCA:CVSL Vrt", vlumContrSS/vlumContr); // CoV

		st.addValue("VCA:MS Hrz", hcolContr); // weighted mean chromatic RNL contrast
		st.addValue("VCA:sS Hrz", hcolContrSS); // weighted SD chromatic RNL contrast
		st.addValue("VCA:CVS Hrz", hcolContrSS/hcolContr); // CoV

		st.addValue("VCA:MS Vrt", vcolContr); // weighted mean chromatic RNL contrast
		st.addValue("VCA:sS Vrt", vcolContrSS); // weighted SD chromatic RNL contrast
		st.addValue("VCA:CVS Vrt", vcolContrSS/vcolContr); // CoV

	}



	//---------------Boundary Strength Analysis Results--------------------


	st.addValue("BSA:BML", meanMichLumT); // Michelson luminance contrast
	st.addValue("BSA:BsL", michLumSST); // Michelson luminance SD
	st.addValue("BSA:BCVL", michLumSST/meanMichLumT); // CoV

	st.addValue("BSA:BMDmax", meanDmaxT); // Dmax Boundary weighted mean
	st.addValue("BSA:BsDmax", DmaxSST); //Dmax Boundary weighted SD
	st.addValue("BSA:BCVDmax", DmaxSST/meanDmaxT); // CoV

	st.addValue("BSA:BMSsat", meanRNLChrT); // RNL Chr Boundary weighted mean overall average chromaticity weighted by image coverage
	st.addValue("BSA:BsSsat", RNLChrSST); //RNL Chr Boundary weighted SD
	st.addValue("BSA:BCVSsat", RNLChrSST/meanRNLChrT); // RNL Chr Boundary CoV

	st.addValue("BSA:BMSL", meanLumT); // LumJND Boundary weighted mean
	st.addValue("BSA:BsSL", LumSST); //SD
	st.addValue("BSA:BCVSL", LumSST/meanLumT); // transition CoV

	st.addValue("BSA:BMS", meanJNDT); // JND Boundary weighted mean
	st.addValue("BSA:BsS", JNDSST); // JND Boundary weighted SD
	st.addValue("BSA:BCVS", JNDSST/meanJNDT); // transition CoV

	if(outputHVdata == true){
		st.addValue("BSA:BML Hrz", hmeanMichLumT); // Michelson luminance contrast
		st.addValue("BSA:BsL Hrz", hmichLumSST); // Michelson luminance SD
		st.addValue("BSA:BCVL Hrz", hmichLumSST/hmeanMichLumT); // CoV

		st.addValue("BSA:BML Vrt", vmeanMichLumT); // Michelson luminance contrast
		st.addValue("BSA:BsL Vrt", vmichLumSST); // Michelson luminance SD
		st.addValue("BSA:BCVL Vrt", vmichLumSST/vmeanMichLumT); // CoV

		st.addValue("BSA:BMDmax Hrz", hmeanDmaxT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsDmax Hrz", hDmaxSST);
		st.addValue("BSA:BCVDmax Hrz", hDmaxSST/hmeanDmaxT); // transition CoV

		st.addValue("BSA:BMDmax Vrt", vmeanDmaxT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsDmax Vrt", vDmaxSST);
		st.addValue("BSA:BCVDmax Vrt", vDmaxSST/vmeanDmaxT); // transition CoV

		st.addValue("BSA:BMSsat Hrz", hmeanRNLChrT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsSsat Hrz", hRNLChrSST);
		st.addValue("BSA:BCVSsat Hrz", hRNLChrSST/hmeanRNLChrT); // transition CoV

		st.addValue("BSA:BMSsat Vrt", vmeanRNLChrT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsSsat Vrt", vRNLChrSST);
		st.addValue("BSA:BCVSsat Vrt", vRNLChrSST/vmeanRNLChrT); // transition CoV

		st.addValue("BSA:BMSL Hrz", hmeanLumT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsSL Hrz", hLumSST);
		st.addValue("BSA:BCVSL Hrz", hLumSST/hmeanLumT); // transition CoV

		st.addValue("BSA:BMSL Vrt", vmeanLumT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsSL Vrt", vLumSST);
		st.addValue("BSA:BCVSL Vrt", vLumSST/vmeanLumT); // transition CoV

		st.addValue("BSA:BMS Hrz", hmeanJNDT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsS Hrz", hJNDSST);
		st.addValue("BSA:BCVS Hrz", hJNDSST/hmeanJNDT); // transition CoV

		st.addValue("BSA:BMS Vrt", vmeanJNDT); // overall average chromaticity weighted by image coverage
		st.addValue("BSA:BsS Vrt", vJNDSST);
		st.addValue("BSA:BCVS Vrt", vJNDSST/vmeanJNDT); // transition CoV

	}

	st.showRowNumbers(false);
	st.show("Results");

	IJ.renameResults("Results", "Summary Results");


IJ.log("_____________________________________________");


}
}
